@@ -48,8 +48,13 @@ logmark_tg(struct sk_buff **pskb, const struct xt_action_param *par)
 	printk(" ctdir=%s", dir_names[ctinfo >= IP_CT_IS_REPLY]);
 	if (ct == NULL) {
 		printk(" ct=NULL ctmark=NULL ctstate=INVALID ctstatus=NONE");
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
+	} else if (nf_ct_is_untracked(ct)) {
+		printk(" ct=UNTRACKED ctmark=NULL ctstate=UNTRACKED ctstatus=NONE");
+#else
 	} else if (ct == &nf_conntrack_untracked) {
 		printk(" ct=UNTRACKED ctmark=NULL ctstate=UNTRACKED ctstatus=NONE");
+#endif
 	} else {
 		printk(" ct=0x%p ctmark=0x%x ctstate=", ct, ct->mark);
 		ctinfo %= IP_CT_IS_REPLY;
